@@ -17,7 +17,7 @@ using namespace apolloron;
 
 const char *INKF_PROGNAME = "inkf";
 const char *INKF_BRIEF    = "Yet Another Network Kanji Filter";
-const char *INKF_VERSION  = "0.9.0";
+const char *INKF_VERSION  = "0.9.1";
 const char *INKF_COPY     = "Copyright (C) 2002-2009 Tachibanamasashi";
 const char *INKF_DEF_OUT  = "inkf.out";
 
@@ -77,7 +77,7 @@ int main(int argc, char *argv[]) {
 
   // set option
   if (set_option(&option, argc, (const char **)argv) == (TOption *)NULL) {
-    fprintf(stderr, "Invalid parameter(s).");
+    fprintf(stderr, "Invalid parameter(s).\n");
     retval = -1;
     goto main_exit;
   }
@@ -543,11 +543,11 @@ void guess(const String &str, const char *input_charset, FILE *fpout) {
     charcode = input_charset;
   }
 
-  retcode = RETURN_STR[LINE_MODE_LF];
+  retcode = "LF";
   if (0 <= str.search(RETURN_STR[LINE_MODE_CRLF])) {
-    retcode = RETURN_STR[LINE_MODE_CRLF];
+    retcode = "CRLF";
   } else if (0 <= str.search(RETURN_STR[LINE_MODE_CR])) {
-    retcode = RETURN_STR[LINE_MODE_CR];
+    retcode = "CR";
   }
 
   fprintf(fpout, "%s (%s)\n", charcode, retcode);
@@ -572,6 +572,8 @@ void show_help() {
       "   SHIFT_JIS (CP932)\n"
       "   EUC-JP\n"
       "   ISO-8859-[1,2,3,4,5,6,7,8,9,10,13,14,15,16]\n"
+      "   KOI8-[R,U]\n"
+      "   CP1251\n"
       "   BIG5 (CP950,EUC-TW)\n"
       "   GB2312 (GBK,GB18030,EUC-CN)\n"
       "   EUC-KR (CP949)\n"
