@@ -24,6 +24,7 @@
 #include "charset/table_unicode_big5.h"
 #include "charset/table_big5_unicode.h"
 #include "charset/table_iso8859_unicode.h"
+#include "charset/table_koi8_unicode.h"
 #include "charset/table_charwidth.h"
 #include "charset.h"
 
@@ -44,6 +45,9 @@ const char *STR_ISO8859_13 = "ISO-8859-13";
 const char *STR_ISO8859_14 = "ISO-8859-14";
 const char *STR_ISO8859_15 = "ISO-8859-15";
 const char *STR_ISO8859_16 = "ISO-8859-16";
+const char *STR_KOI8_R     = "KOI8-R";
+const char *STR_KOI8_U     = "KOI8-U";
+const char *STR_CP1251     = "CP1251";
 const char *STR_UTF8       = "UTF-8";
 const char *STR_UTF7       = "UTF-7";
 const char *STR_UTF7_IMAP  = "UTF-7-IMAP";
@@ -89,7 +93,8 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
   } else if (!strncasecmp(src_charset, "EUC-JP", 6) || !strncasecmp(src_charset, "EUCJP", 5)) {
     src_ch = STR_EUCJP;
   } else if (!strcasecmp(src_charset, "SHIFT_JIS") || !strcasecmp(src_charset, "SHIFT-JIS") ||
-      !strcasecmp(src_charset, "CP932") || !strcasecmp(src_charset, "MS932")) {
+      !strcasecmp(src_charset, "CP932") || !strcasecmp(src_charset, "WINDOWS-932") ||
+      !strcasecmp(src_charset, "MS932")) {
     src_ch = STR_SJIS;
   } else if (!strncasecmp(src_charset, "ISO-2022-JP", 11)) {
     src_ch = STR_JIS;
@@ -110,13 +115,22 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
       src_ch = src_charset;
     }
   } else if (!strncasecmp(src_charset, "EUC-KR", 6) || !strncasecmp(src_charset, "EUCKR", 5) ||
-      !strcasecmp(src_charset, "CP949") || !strcasecmp(src_charset, "MS949")) {
+      !strcasecmp(src_charset, "CP949") || !strcasecmp(src_charset, "WINDOWS-949") ||
+      !strcasecmp(src_charset, "MS949")) {
     src_ch = STR_EUCKR;
   } else if (!strncasecmp(src_charset, "GB", 2) || !strncasecmp(src_charset, "EUC-CN", 6) || !strncasecmp(src_charset, "EUCCN", 5)) {
     src_ch = STR_GBK;
   } else if (!strcasecmp(src_charset, "BIG5") || !strncasecmp(src_charset, "EUC-TW", 6) ||
-    !strncasecmp(src_charset, "EUCTW", 5) || !strcasecmp(src_charset, "CP950") || !strcasecmp(src_charset, "MS950")) {
+    !strncasecmp(src_charset, "EUCTW", 5) || !strcasecmp(src_charset, "CP950") ||
+    !strcasecmp(src_charset, "WINDOWS-950") || !strcasecmp(src_charset, "MS950")) {
     src_ch = STR_BIG5;
+  } else if (!strcasecmp(src_charset, "CP1251") || !strcasecmp(src_charset, "WINDOWS-1251") ||
+      !strcasecmp(src_charset, "MS1251")) {
+    src_ch = STR_CP1251;
+  } else if (!strncasecmp(src_charset, "KOI8-U", 6)) {
+    src_ch = STR_KOI8_U;
+  } else if (!strncasecmp(src_charset, "KOI8", 4)) {
+    src_ch = STR_KOI8_R;
   } else {
     src_ch = src_charset;
   }
@@ -126,7 +140,8 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
   } else if (!strncasecmp(dest_charset, "EUC-JP", 6) || !strncasecmp(dest_charset, "EUCJP", 5)) {
     dest_ch = STR_EUCJP;
   } else if (!strcasecmp(dest_charset, "SHIFT_JIS") || !strcasecmp(dest_charset, "SHIFT-JIS") ||
-      !strcasecmp(dest_charset, "CP932") || !strcasecmp(dest_charset, "MS932")) {
+      !strcasecmp(dest_charset, "CP932") || !strcasecmp(dest_charset, "WINDOWS-932") ||
+      !strcasecmp(dest_charset, "MS932")) {
     dest_ch = STR_SJIS;
   } else if (!strncasecmp(dest_charset, "ISO-2022-JP", 11)) {
     dest_ch = STR_JIS;
@@ -137,13 +152,22 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
   } else if (!strncasecmp(dest_charset, "UTF-7", 5) || !strncasecmp(dest_charset, "UTF7", 4)) {
     dest_ch = STR_UTF7;
   } else if (!strncasecmp(dest_charset, "EUC-KR", 6) || !strncasecmp(dest_charset, "EUCKR", 5) ||
-      !strcasecmp(dest_charset, "CP949") || !strcasecmp(dest_charset, "MS949")) {
+      !strcasecmp(dest_charset, "CP949") || !strcasecmp(dest_charset, "WINDOWS-949") ||
+      !strcasecmp(dest_charset, "MS949")) {
     dest_ch = STR_EUCKR;
   } else if (!strncasecmp(dest_charset, "GB", 2) || !strncasecmp(dest_charset, "EUC-CN", 6) || !strncasecmp(src_charset, "EUCCN", 5)) {
     dest_ch = STR_GBK;
   } else if (!strcasecmp(dest_charset, "BIG5") || !strncasecmp(dest_charset, "EUC-TW", 6) ||
-    !strncasecmp(dest_charset, "EUCTW", 5) || !strcasecmp(dest_charset, "CP950") || !strcasecmp(dest_charset, "MS950")) {
+    !strncasecmp(dest_charset, "EUCTW", 5) || !strcasecmp(dest_charset, "CP950") ||
+    !strcasecmp(dest_charset, "WINDOWS-950") || !strcasecmp(dest_charset, "MS950")) {
     dest_ch = STR_BIG5;
+  } else if (!strcasecmp(dest_charset, "CP1251") || !strcasecmp(dest_charset, "WINDOWS-1251") ||
+      !strcasecmp(dest_charset, "MS1251")) {
+    dest_ch = STR_CP1251;
+  } else if (!strncasecmp(dest_charset, "KOI8-U", 6)) {
+    dest_ch = STR_KOI8_U;
+  } else if (!strncasecmp(dest_charset, "KOI8", 4)) {
+    dest_ch = STR_KOI8_R;
   } else {
     dest_ch = "";
   }
@@ -289,6 +313,15 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
       num = atoi(dest_ch + 9);
       buf = utf8_to_iso8859(str, num);
       converted = 1;
+    } else if (!strcasecmp(dest_ch, "KOI8-U")) {
+      buf = utf8_to_koi8(str, 'U');
+      converted = 1;
+    } else if (!strcasecmp(dest_ch, "KOI8-R")) {
+      buf = utf8_to_koi8(str, 'R');
+      converted = 1;
+    } else if (!strcasecmp(dest_ch, "CP1251")) {
+      buf = utf8_to_koi8(str, 'W');
+      converted = 1;
     }
   } else if (!strcasecmp(src_ch, STR_AUTOJP) || !strcasecmp(src_ch, STR_AUTO)) {
     if (!strcasecmp(dest_ch, STR_EUCJP)) {
@@ -341,7 +374,22 @@ char* charset_convert(const char* str, const char *src_charset, const char *dest
     if (!strcasecmp(dest_ch, STR_UTF8)) {
       int num;
       num = atoi(dest_ch + 9);
-      buf = utf8_to_iso8859(str, num);
+      buf = iso8859_to_utf8(str, num);
+      converted = 1;
+    }
+  } else if (!strcasecmp(src_ch, "KOI8-U")) {
+    if (!strcasecmp(dest_ch, STR_UTF8)) {
+      buf = koi8_to_utf8(str, 'U');
+      converted = 1;
+    }
+  } else if (!strcasecmp(src_ch, "KOI8-R")) {
+    if (!strcasecmp(dest_ch, STR_UTF8)) {
+      buf = koi8_to_utf8(str, 'R');
+      converted = 1;
+    }
+  } else if (!strcasecmp(src_ch, "CP1251")) {
+    if (!strcasecmp(dest_ch, STR_UTF8)) {
+      buf = koi8_to_utf8(str, 'W');
       converted = 1;
     }
   }
@@ -582,7 +630,7 @@ const char* jis_auto_detect(const char *str) {
 
 #if __ICONV == 0
 /*! Convert character set from ISO-8859-x to UTF-8
-    @param str          ISO-8859-x text.
+    @param str          ISO-8859-x text
     @param iso8859_num  x of ISO-8859-x.
     @return Converted text  (UTF-8 text)
  */
@@ -590,54 +638,57 @@ char* iso8859_to_utf8(const char* str, int iso8859_num) {
   char *buf;
   long i, j, length, length2;
   unsigned char a0, a1;
-  const unsigned char ***iso8859_map;
+  const unsigned char (*iso8859_map)[ISO8859_UNICODE_TABLE_MAX][2];
 
   if (iso8859_num <= 0 || (11 <= iso8859_num && iso8859_num <= 12) || 17 <= iso8859_num) {
     return NULL;
   }
 
   switch (iso8859_num) {
+    case 1:
+        iso8859_map = &iso8859_1_unicode;
+        break;
     case 2:
-        iso8859_map = (const unsigned char ***)&iso8859_2_unicode;
+        iso8859_map = &iso8859_2_unicode;
         break;
     case 3:
-        iso8859_map = (const unsigned char ***)&iso8859_3_unicode;
+        iso8859_map = &iso8859_3_unicode;
         break;
     case 4:
-        iso8859_map = (const unsigned char ***)&iso8859_4_unicode;
+        iso8859_map = &iso8859_4_unicode;
         break;
     case 5:
-        iso8859_map = (const unsigned char ***)&iso8859_5_unicode;
+        iso8859_map = &iso8859_5_unicode;
         break;
     case 6:
-        iso8859_map = (const unsigned char ***)&iso8859_6_unicode;
+        iso8859_map = &iso8859_6_unicode;
         break;
     case 7:
-        iso8859_map = (const unsigned char ***)&iso8859_7_unicode;
+        iso8859_map = &iso8859_7_unicode;
         break;
     case 8:
-        iso8859_map = (const unsigned char ***)&iso8859_8_unicode;
+        iso8859_map = &iso8859_8_unicode;
         break;
     case 9:
-        iso8859_map = (const unsigned char ***)&iso8859_9_unicode;
+        iso8859_map = &iso8859_9_unicode;
         break;
     case 10:
-        iso8859_map = (const unsigned char ***)&iso8859_10_unicode;
+        iso8859_map = &iso8859_10_unicode;
         break;
     case 13:
-        iso8859_map = (const unsigned char ***)&iso8859_13_unicode;
+        iso8859_map = &iso8859_13_unicode;
         break;
     case 14:
-        iso8859_map = (const unsigned char ***)&iso8859_14_unicode;
+        iso8859_map = &iso8859_14_unicode;
         break;
     case 15:
-        iso8859_map = (const unsigned char ***)&iso8859_15_unicode;
+        iso8859_map = &iso8859_15_unicode;
         break;
     case 16:
-        iso8859_map = (const unsigned char ***)&iso8859_16_unicode;
+        iso8859_map = &iso8859_16_unicode;
         break;
     default:
-        iso8859_map = (const unsigned char ***)0;
+        return NULL;
         break;
   }
 
@@ -654,16 +705,12 @@ char* iso8859_to_utf8(const char* str, int iso8859_num) {
     j = 0;
     while (i < length) {
       if ((unsigned char)0x80 <= (unsigned char)str[i]) {
-        if (iso8859_map == (const unsigned char ***)0) {
-          /* ISO-8859-1 */
-          a0 = 0;
-          a1 = str[i];
-        } else {
-          a0 = (*iso8859_map)[(unsigned char)str[i] - (unsigned char)0x80][1];
-          a1 = (*iso8859_map)[(unsigned char)str[i] - (unsigned char)0x80][2];
-        }
-        if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
-          buf[j++] = 0xC0 | (a1 >> 6);
+        a0 = (*iso8859_map)[(unsigned char)str[i] - (unsigned char)0x80][0];
+        a1 = (*iso8859_map)[(unsigned char)str[i] - (unsigned char)0x80][1];
+	if (a0 == (unsigned char)0x00) {
+          buf[j++] = a1;
+	} else if (a0 <= (unsigned char)0x07) {
+          buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
           buf[j++] = 0x80 | (0x3F & a1);
         } else {
           buf[j++] = 0xE0 | (a0 >> 4);
@@ -671,7 +718,7 @@ char* iso8859_to_utf8(const char* str, int iso8859_num) {
           buf[j++] = 0x80 | (0x3F & a1);
         }
       } else {
-        buf[j++] = '?';
+        buf[j++] = str[i];
       }
       i++;
     }
@@ -691,54 +738,57 @@ char* utf8_to_iso8859(const char* str, int iso8859_num) {
   char *buf;
   long i, j, length, length2;
   unsigned char a0, a1, a2;
-  const unsigned char ***iso8859_map;
+  const unsigned char (*iso8859_map)[ISO8859_UNICODE_TABLE_MAX][2];
 
   if (iso8859_num <= 0 || (11 <= iso8859_num && iso8859_num <= 12) || 17 <= iso8859_num) {
     return NULL;
   }
 
   switch (iso8859_num) {
+    case 1:
+        iso8859_map = &iso8859_1_unicode;
+        break;
     case 2:
-        iso8859_map = (const unsigned char ***)&iso8859_2_unicode;
+        iso8859_map = &iso8859_2_unicode;
         break;
     case 3:
-        iso8859_map = (const unsigned char ***)&iso8859_3_unicode;
+        iso8859_map = &iso8859_3_unicode;
         break;
     case 4:
-        iso8859_map = (const unsigned char ***)&iso8859_4_unicode;
+        iso8859_map = &iso8859_4_unicode;
         break;
     case 5:
-        iso8859_map = (const unsigned char ***)&iso8859_5_unicode;
+        iso8859_map = &iso8859_5_unicode;
         break;
     case 6:
-        iso8859_map = (const unsigned char ***)&iso8859_6_unicode;
+        iso8859_map = &iso8859_6_unicode;
         break;
     case 7:
-        iso8859_map = (const unsigned char ***)&iso8859_7_unicode;
+        iso8859_map = &iso8859_7_unicode;
         break;
     case 8:
-        iso8859_map = (const unsigned char ***)&iso8859_8_unicode;
+        iso8859_map = &iso8859_8_unicode;
         break;
     case 9:
-        iso8859_map = (const unsigned char ***)&iso8859_9_unicode;
+        iso8859_map = &iso8859_9_unicode;
         break;
     case 10:
-        iso8859_map = (const unsigned char ***)&iso8859_10_unicode;
+        iso8859_map = &iso8859_10_unicode;
         break;
     case 13:
-        iso8859_map = (const unsigned char ***)&iso8859_13_unicode;
+        iso8859_map = &iso8859_13_unicode;
         break;
     case 14:
-        iso8859_map = (const unsigned char ***)&iso8859_14_unicode;
+        iso8859_map = &iso8859_14_unicode;
         break;
     case 15:
-        iso8859_map = (const unsigned char ***)&iso8859_15_unicode;
+        iso8859_map = &iso8859_15_unicode;
         break;
     case 16:
-        iso8859_map = (const unsigned char ***)&iso8859_16_unicode;
+        iso8859_map = &iso8859_16_unicode;
         break;
     default:
-        iso8859_map = (const unsigned char ***)0;
+        return NULL;
         break;
   }
 
@@ -778,6 +828,10 @@ char* utf8_to_iso8859(const char* str, int iso8859_num) {
         } else if (len1 == 3) {
           a0 = (str[i] << 4) | (0x0F & (str[i+1] >> 2));
           a1 = (str[i+1] << 6) | (0x3F & str[i+2]);
+        } else if (len1 == 1) {
+          buf[j++] = str[i];
+          i++;
+          continue;
         } else {
           buf[j++] = '?';
           i += len1;
@@ -808,6 +862,150 @@ char* utf8_to_iso8859(const char* str, int iso8859_num) {
 }
 
 
+/*! Convert character set from KOI8-x to UTF-8
+    @param str          KOI-8-x text
+    @param ch           x of KOI8-x. 'w' for CP1251
+    @return Converted text  (UTF-8 text)
+ */
+char* koi8_to_utf8(const char* str, char ch) {
+  char *buf;
+  long i, j, length, length2;
+  unsigned char a0, a1;
+  const unsigned char (*koi8_map)[KOI8_UNICODE_TABLE_MAX][2];
+
+  if (ch == 'W' || ch == 'w') {
+    koi8_map = &cp1251_unicode;
+  } else if (ch == 'U' || ch == 'u') {
+    koi8_map = &koi8_u_unicode;
+  } else {
+    koi8_map = &koi8_r_unicode;
+  }
+
+  buf = NULL;
+
+  if (str != NULL) {
+    // Memory allocation
+    length = strlen(str);
+    length2 = length*3;
+    buf = new char[length2 + 1];
+
+    // KOI8-x to UTF-8
+    i = 0;
+    j = 0;
+    while (i < length) {
+      if ((unsigned char)0x80 <= (unsigned char)str[i]) {
+        a0 = (*koi8_map)[(unsigned char)str[i] - (unsigned char)0x80][0];
+        a1 = (*koi8_map)[(unsigned char)str[i] - (unsigned char)0x80][1];
+	if (a0 == (unsigned char)0x00) {
+          buf[j++] = a1;
+	} else if (a0 <= (unsigned char)0x07) {
+          buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
+          buf[j++] = 0x80 | (0x3F & a1);
+        } else {
+          buf[j++] = 0xE0 | (a0 >> 4);
+          buf[j++] = 0x80 | ((0x3F & (a0 << 2)) | (a1 >> 6));
+          buf[j++] = 0x80 | (0x3F & a1);
+        }
+      } else {
+        buf[j++] = str[i];
+      }
+      i++;
+    }
+    buf[j] = '\0';
+  }
+
+  return buf;
+}
+
+
+/*! Convert character set from UTF-8 to KOI8-x
+    @param str          UTF-8 text
+    @param ch           x of KOI8-x. 'w' for CP1251
+    @return Converted text (KOI8-x text)
+ */
+char* utf8_to_koi8(const char* str, char ch) {
+  char *buf;
+  long i, j, length, length2;
+  unsigned char a0, a1, a2;
+  const unsigned char (*koi8_map)[KOI8_UNICODE_TABLE_MAX][2];
+
+  if (ch == 'W' || ch == 'w') {
+    koi8_map = &cp1251_unicode;
+  } else if (ch == 'U' || ch == 'u') {
+    koi8_map = &koi8_u_unicode;
+  } else {
+    koi8_map = &koi8_r_unicode;
+  }
+
+  buf = NULL;
+
+  if (str != NULL) {
+    // Memory allocation
+    length = strlen(str);
+    length2 = length*3;
+    buf = new char[length2 + 1];
+
+    // UTF-8 to KOI8-x
+    i = 0;
+    j = 0;
+    while (i < length && j <= length2) {
+      a0 = str[i];
+      if (isutf8_1(a0) && i+utf8_len(a0)-1 < length) {
+        unsigned short len1, n;
+
+        a1 = str[i+1];
+        a2 = str[i+2];
+        if (a0 == 0xEF && a1 == 0xBB && a2 == 0xBF) {
+          i += 3;
+          continue;
+        }
+
+        if (a0 == 0xE2 && a1 == 0x80 && a2 == 0xBE) {
+          buf[j++] = 0x7E;
+          i += 3;
+          continue;
+        }
+
+        len1 = utf8_len(str[i]);
+        if (len1 == 2) {
+          a0 = 0x07 & (str[i] >> 2);
+          a1 = (str[i] << 6) | (0x3F & str[i+1]);
+        } else if (len1 == 3) {
+          a0 = (str[i] << 4) | (0x0F & (str[i+1] >> 2));
+          a1 = (str[i+1] << 6) | (0x3F & str[i+2]);
+        } else if (len1 == 1) {
+          buf[j++] = str[i];
+          i++;
+          continue;
+        } else {
+          buf[j++] = '?';
+          i += len1;
+          continue;
+        }
+
+        n = 0;
+        while (n < KOI8_UNICODE_TABLE_MAX) {
+          if ((*koi8_map)[n][0] == a0 && (*koi8_map)[n][1] == a1) {
+            buf[j++] = (unsigned char)(0x80 + n);
+            i += (utf8_len(str[i])-1);
+            break;
+          }
+          n++;
+        }
+        if (KOI8_UNICODE_TABLE_MAX <= n) {
+          buf[j++] = '?';
+        }
+      } else {
+        buf[j++] = str[i];
+      }
+      i++;
+    }
+    buf[j] = '\0';
+  }
+  return buf;
+}
+
+
 /*! Convert character set from UTF-8 to EUC-JP
     @param str  UTF-8 text
     @return Converted text (EUC-JP text)
@@ -1328,8 +1526,10 @@ char* sjis_to_utf8(const char *str) {
         if (0 <= min && pos != old_pos && min <= max) {
           a0 = sjis_unicode[pos][1];
           a1 = sjis_unicode[pos][2];
-          if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
-            buf[j++] = 0xC0 | (a1 >> 6);
+	  if (a0 == (unsigned char)0x00) {
+            buf[j++] = a1;
+	  } else if (a0 <= (unsigned char)0x07) {
+            buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
             buf[j++] = 0x80 | (0x3F & a1);
           } else {
             buf[j++] = 0xE0 | (a0 >> 4);
@@ -1782,8 +1982,10 @@ char* euckr_to_utf8(const char* str) {
         if (0 <= min && pos != old_pos && min <= max) {
           a0 = euckr_unicode[pos][1];
           a1 = euckr_unicode[pos][2];
-          if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
-            buf[j++] = 0xC0 | (a1 >> 6);
+	  if (a0 == (unsigned char)0x00) {
+            buf[j++] = a1;
+	  } else if (a0 <= (unsigned char)0x07) {
+            buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
             buf[j++] = 0x80 | (0x3F & a1);
           } else {
             buf[j++] = 0xE0 | (a0 >> 4);
@@ -1960,8 +2162,10 @@ char* gbk_to_utf8(const char* str) {
           a0 = gbk_unicode[pos][0];
           a1 = gbk_unicode[pos][1];
           if (!(a0 == 0 && a1 == 0)) {
-            if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
-              buf[j++] = 0xC0 | (a1 >> 6);
+	    if (a0 == (unsigned char)0x00) {
+              buf[j++] = a1;
+	    } else if (a0 <= (unsigned char)0x07) {
+              buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
               buf[j++] = 0x80 | (0x3F & a1);
             } else {
               buf[j++] = 0xE0 | (a0 >> 4);
@@ -2145,8 +2349,10 @@ char* big5_to_utf8(const char* str) {
         if (0 <= min && pos != old_pos && min <= max) {
           a0 = big5_unicode[pos][1];
           a1 = big5_unicode[pos][2];
-          if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
-            buf[j++] = 0xC0 | (a1 >> 6);
+	  if (a0 == (unsigned char)0x00) {
+            buf[j++] = a1;
+	  } else if (a0 <= (unsigned char)0x07) {
+            buf[j++] = 0xC0 | ((a0 & 7) << 2) | (a1 >> 6);
             buf[j++] = 0x80 | (0x3F & a1);
           } else {
             buf[j++] = 0xE0 | (a0 >> 4);
@@ -2561,10 +2767,15 @@ char* modutf7_to_utf8(const char* str) {
           delete [] utf8_str;
           utf8_str = tmp_utf8_str;
         }
-        utf8_str[col] = '%';
-        utf8_str[col+1] = hex[c >> 4];
-        utf8_str[col+2] = hex[c & 0x0f];
-        col += 3;
+        if (c != '\r' && c != '\n') {
+          utf8_str[col] = '%';
+          utf8_str[col+1] = hex[c >> 4];
+          utf8_str[col+2] = hex[c & 0x0f];
+          col += 3;
+        } else {
+          utf8_str[col] = (char)c;
+          col++;
+        }
       } else {
         // encode literally
         if (str_malloced_size - col - 1 < 1) {
